﻿using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.SharePoint;
using System.Collections.ObjectModel;
using Microsoft.SharePoint.Publishing;
using Microsoft.SharePoint.Utilities;
using System.IO;
using System.Xml;
using System.Security.Cryptography;
using System.Web.UI.WebControls.WebParts;
using System.Text.RegularExpressions;
using System.Web;

namespace WCMexport
{
    public class WCMexportLib : IDisposable
    {
        SPSite _Site = null;
        PublishingSite _PubSite = null;
        SPWeb _RootWeb = null;

        public WCMexportLib(string url)
        {
            _Site = new SPSite(url);
            _PubSite = new PublishingSite(_Site);
            _RootWeb = _Site.RootWeb;
        }

        #region public ReadOnlyCollection<UrlInfo> GetMasterPages()
        public ReadOnlyCollection<UrlInfo> GetMasterPages()
        {
            SPList mpc = _RootWeb.GetCatalog(SPListTemplateType.MasterPageCatalog);
            StringBuilder sbQuery=new StringBuilder();
            sbQuery.Append("<Where><And>");
            sbQuery.AppendFormat("<Neq><FieldRef Name='{0}'/><Value Type='Boolean'>1</Value></Neq>", SPHttpUtility.HtmlEncode(mpc.Fields[FieldId.Hidden].InternalName));
            sbQuery.AppendFormat("<Eq><FieldRef Name='{0}'/><Value Type='Text'>master</Value></Eq>", SPHttpUtility.HtmlEncode(mpc.Fields[SPBuiltInFieldId.File_x0020_Type].InternalName));
            sbQuery.Append("</And></Where>");
            sbQuery.AppendFormat("<OrderBy><FieldRef Name=\"{0}\" /><FieldRef Name=\"{1}\" /></OrderBy>", SPHttpUtility.HtmlEncode(mpc.Fields[SPBuiltInFieldId.FileDirRef].InternalName),SPHttpUtility.HtmlEncode(mpc.Fields[FieldId.Title].InternalName));
            SPQuery query = new SPQuery();
            query.ViewAttributes = "Scope=\"Recursive\"";
            query.RowLimit = 0;
            query.Query = sbQuery.ToString();
            List<UrlInfo> result = new List<UrlInfo>();
            foreach (SPListItem item in mpc.GetItems(query))
            {
                result.Add(new UrlInfo((string)item[SPBuiltInFieldId.FileLeafRef], item.File.ServerRelativeUrl, (string)item["MasterPageDescription"]));
            }
            return new ReadOnlyCollection<UrlInfo>(result);
        }
        #endregion

        #region public ReadOnlyCollection<UrlInfo> GetPageLayouts()
        public ReadOnlyCollection<UrlInfo> GetPageLayouts()
        {
            List<UrlInfo> result = new List<UrlInfo>();
            foreach (PageLayout pl in _PubSite.PageLayouts)
            {
                result.Add(new UrlInfo(pl.Title, pl.ServerRelativeUrl, pl.Description));
            }
            return new ReadOnlyCollection<UrlInfo>(result);
        }
        #endregion

        #region public ReadOnlyCollection<UrlInfo> GetStyleLibraryFiles()
        public ReadOnlyCollection<UrlInfo> GetStyleLibraryFiles()
        {
            SPList mpc = _RootWeb.GetList("Style Library");
            StringBuilder sbQuery = new StringBuilder();
            sbQuery.AppendFormat("<OrderBy><FieldRef Name=\"{0}\" /><FieldRef Name=\"{1}\" /></OrderBy>", SPHttpUtility.HtmlEncode(mpc.Fields[SPBuiltInFieldId.FileDirRef].InternalName), SPHttpUtility.HtmlEncode(mpc.Fields[SPBuiltInFieldId.FileLeafRef].InternalName));
            SPQuery query = new SPQuery();
            query.ViewAttributes = "Scope=\"Recursive\"";
            query.RowLimit = 0;
            query.Query = sbQuery.ToString();
            List<UrlInfo> result = new List<UrlInfo>();
            foreach (SPListItem item in mpc.GetItems(query))
            {
                result.Add(new UrlInfo((string)item[SPBuiltInFieldId.FileLeafRef], item.File.ServerRelativeUrl, String.Format("Created By {0} on {1}", GetFieldValue<SPFieldUserValue>(item, SPBuiltInFieldId.Author).LookupValue, item.File.TimeCreated)));
            }
            return new ReadOnlyCollection<UrlInfo>(result);
        }
        #endregion

        public Exporter StartExport(string path)
        {
            return new Exporter(this, path);
        }

        #region Utility Methods
        internal static T GetFieldValue<T>(SPListItem item, Guid fieldID)
        {
            SPField field = item.Fields[fieldID];
            return (T)field.GetFieldValue(item[field.Id].ToString());
        }

        #region Dispose
        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                if (_RootWeb != null)
                {
                    _RootWeb.Dispose();
                    _RootWeb = null;
                }
                if (_Site != null)
                {
                    _Site.Dispose();
                    _Site = null;
                }
            }
        }
        #endregion
        #endregion

        public class Exporter : IDisposable
        {
            private const string nsSharePoint = "http://schemas.microsoft.com/sharepoint/";
            WCMexportLib _WCMexportLib;
            string _Dir;
            XmlTextWriter _FeatureXml;
            XmlTextWriter _MasterPagesXml;
            Dictionary<string, bool> _PreviousPreviewUrls = new Dictionary<string, bool>();
            List<string> _MasterPagePreviewUrls = new List<string>();
            XmlTextWriter _PageLayoutsXml;
            List<string> _PageLayoutPreviewUrls = new List<string>();
            Dictionary<string, string> _PageLayoutFileNames = new Dictionary<string, string>();
            string _CurrentStyleLibrarySubdir;
            XmlTextWriter _StyleLibraryFilesXml;
            internal Exporter(WCMexportLib wcmExportLib, string path)
            {
                _WCMexportLib = wcmExportLib;
                _Dir = Path.GetDirectoryName(path);
                _FeatureXml = new XmlTextWriter(path, Encoding.UTF8);
                _FeatureXml.Formatting = Formatting.Indented;
                _FeatureXml.WriteStartElement("Feature",nsSharePoint);
                _FeatureXml.WriteAttributeString("Id",Guid.NewGuid().ToString("b").Substring(1,36));
                _FeatureXml.WriteAttributeString("Title", String.Format("WCM export from {0}", _WCMexportLib._RootWeb.Title));
                _FeatureXml.WriteAttributeString("Description", String.Format("WCM export from {0}", _WCMexportLib._RootWeb.Title));
                _FeatureXml.WriteAttributeString("Scope","Site");
                _FeatureXml.WriteStartElement("ElementManifests",nsSharePoint);
            }

            public void AddMasterPage(string url)
            {
                if (_MasterPagesXml == null)
                {
                    AddElementManifest("MasterPages.xml");

                    _MasterPagesXml = new XmlTextWriter(Path.Combine(_Dir, "MasterPages.xml"), Encoding.UTF8);
                    _MasterPagesXml.Formatting = Formatting.Indented;
                    _MasterPagesXml.WriteStartElement("Elements", nsSharePoint);
                    _MasterPagesXml.WriteStartElement("Module", nsSharePoint);
                    _MasterPagesXml.WriteAttributeString("Name", "MasterPages");
                    _MasterPagesXml.WriteAttributeString("Url", "_catalogs/masterpage");
                    _MasterPagesXml.WriteAttributeString("List", "116");
                    _MasterPagesXml.WriteAttributeString("Path", "MasterPages");
                    _MasterPagesXml.WriteAttributeString("RootWebOnly","TRUE");
                    Directory.CreateDirectory(Path.Combine(_Dir, "MasterPages"));
                }
                SPFile file = _WCMexportLib._RootWeb.GetFile(url);
                PageLayout pl = new PageLayout(file.Item);
                string fileName = RemoveUnderscore(pl.Name);
                WriteToFile(file, Path.Combine(Path.Combine(_Dir, "MasterPages"), fileName));
                AddElementFile(Path.Combine("MasterPages", fileName));

                _MasterPagesXml.WriteStartElement("File");
                _MasterPagesXml.WriteAttributeString("Url", fileName);
                _MasterPagesXml.WriteAttributeString("Type","GhostableInLibrary");

                WriteProperty(_MasterPagesXml, "ContentType", "$Resources:cmscore,contenttype_masterpage_name;");
                WriteProperty(_MasterPagesXml, "MasterPageDescription", RemoveUnderscore(pl.Description));

                if (!string.IsNullOrEmpty(pl.PreviewImageUrl))
                {
                    WriteProperty(_MasterPagesXml, "PublishingPreviewImage", String.Format("{0}, {0}", HandlePreview(_MasterPagePreviewUrls, pl.PreviewImageUrl)));
                }
                _MasterPagesXml.WriteEndElement();
            }

            public void AddPageLayout(string url)
            {
                if (_PageLayoutsXml == null)
                {
                    AddElementManifest("PageLayouts.xml");

                    _PageLayoutsXml = new XmlTextWriter(Path.Combine(_Dir, "PageLayouts.xml"), Encoding.UTF8);
                    _PageLayoutsXml.Formatting = Formatting.Indented;
                    _PageLayoutsXml.WriteStartElement("Elements", nsSharePoint);
                    _PageLayoutsXml.WriteStartElement("Module", nsSharePoint);
                    _PageLayoutsXml.WriteAttributeString("Name", "PageLayouts");
                    _PageLayoutsXml.WriteAttributeString("Url", "_catalogs/masterpage");
                    _PageLayoutsXml.WriteAttributeString("List", "116");
                    _PageLayoutsXml.WriteAttributeString("Path", "MasterPages");
                    _PageLayoutsXml.WriteAttributeString("RootWebOnly", "TRUE");
                    Directory.CreateDirectory(Path.Combine(_Dir, "MasterPages"));
                }
                SPFile file = _WCMexportLib._RootWeb.GetFile(url);
                PageLayout pl = new PageLayout(file.Item);
                string fileName = RemoveUnderscore(pl.Name);

                // Check if the PageLayout is the same as one of the previous (probably 
                // then with different ContentType or WebParts)
                //
                string key = GetMD5hash(file);
                string existingFileName;
                if (!_PageLayoutFileNames.TryGetValue(key, out existingFileName))
                {
                    _PageLayoutFileNames.Add(key, fileName);
                    WriteToFile(file, Path.Combine(Path.Combine(_Dir, "MasterPages"), fileName));
                    AddElementFile(Path.Combine("MasterPages", fileName));
                }

                _PageLayoutsXml.WriteStartElement("File");
                _PageLayoutsXml.WriteAttributeString("Url", fileName);
                if (!string.IsNullOrEmpty(existingFileName))
                    _PageLayoutsXml.WriteAttributeString("Path", existingFileName);
                _PageLayoutsXml.WriteAttributeString("Type", "GhostableInLibrary");

                WriteProperty(_PageLayoutsXml, "ContentType", "$Resources:cmscore,contenttype_pagelayout_name;");
                WriteProperty(_PageLayoutsXml, "Title", RemoveUnderscore(pl.Title));
                WriteProperty(_PageLayoutsXml, "MasterPageDescription", RemoveUnderscore(pl.Description));
                WriteProperty(_PageLayoutsXml, "PublishingAssociatedContentType", String.Format(";#{0};#{1};#", pl.AssociatedContentType.Name, pl.AssociatedContentType.Id.ToString()));
                if (!string.IsNullOrEmpty(pl.PreviewImageUrl))
                {
                    WriteProperty(_PageLayoutsXml, "PublishingPreviewImage", String.Format("{0}, {0}", HandlePreview(_PageLayoutPreviewUrls, pl.PreviewImageUrl)));
                }

                Microsoft.SharePoint.WebPartPages.SPLimitedWebPartManager lwpm = file.GetLimitedWebPartManager(PersonalizationScope.Shared);
                foreach (WebPart wp in lwpm.WebParts)
                {
                    string wpExport = GetWebPartXml(lwpm, wp);
                    _PageLayoutsXml.WriteStartElement("AllUsersWebPart", nsSharePoint);
                    _PageLayoutsXml.WriteAttributeString("WebPartZoneID", lwpm.GetZoneID(wp));
                    // Multiply ZoneIndex by 2 so we get Order 0,2,4,...
                    // As sharepoint sometimes get the order wrong with 0,1,2,...
                    _PageLayoutsXml.WriteAttributeString("WebPartOrder", (wp.ZoneIndex * 2).ToString());
                    _PageLayoutsXml.WriteCData(wpExport);
                    _PageLayoutsXml.WriteEndElement();
                }
                _PageLayoutsXml.WriteEndElement();
            }

            public void AddStyleLibraryFile(string url)
            {
                if (_StyleLibraryFilesXml == null)
                {
                    AddElementManifest("StyleLibrary.xml");

                    _StyleLibraryFilesXml = new XmlTextWriter(Path.Combine(_Dir, "StyleLibrary.xml"), Encoding.UTF8);
                    _StyleLibraryFilesXml.Formatting = Formatting.Indented;
                    _StyleLibraryFilesXml.WriteStartElement("Elements", nsSharePoint);
                }

                SPFile file = _WCMexportLib._RootWeb.GetFile(url);

                string subdir = GetSubDirOf("Style Library",file.ServerRelativeUrl);

                if (_CurrentStyleLibrarySubdir==null || _CurrentStyleLibrarySubdir!=subdir)
                {
                    if (_CurrentStyleLibrarySubdir!=null)
                        _StyleLibraryFilesXml.WriteEndElement();
                    _CurrentStyleLibrarySubdir = subdir;

                    _StyleLibraryFilesXml.WriteStartElement("Module", nsSharePoint);
                    _StyleLibraryFilesXml.WriteAttributeString("Name", "StyleLibrary"+subdir);
                    _StyleLibraryFilesXml.WriteAttributeString("Url", "Style Library"+subdir);
                    _StyleLibraryFilesXml.WriteAttributeString("Path", "StyleLibrary"+subdir);
                    _StyleLibraryFilesXml.WriteAttributeString("RootWebOnly", "TRUE");
                    Directory.CreateDirectory(Path.Combine(_Dir, "StyleLibrary" + subdir));
                }

                PageLayout pl = new PageLayout(file.Item);
                string fileName = RemoveUnderscore(pl.Name);
                WriteToFile(file, Path.Combine(Path.Combine(_Dir, "StyleLibrary"+subdir), fileName));
                AddElementFile(Path.Combine("StyleLibrary"+subdir, fileName));

                _StyleLibraryFilesXml.WriteStartElement("File");
                _StyleLibraryFilesXml.WriteAttributeString("Url", fileName);
                _StyleLibraryFilesXml.WriteAttributeString("Type", "GhostableInLibrary");
                _StyleLibraryFilesXml.WriteEndElement();
            }

            private string GetSubDirOf(string rootDir, string fullPath)
            {
                string relativePath = fullPath.Substring(fullPath.IndexOf(rootDir) + rootDir.Length);
                List<string> parts = new List<string>(relativePath.Split(new char[] { '/' }));
                parts.RemoveAt(parts.Count-1);
                return string.Join("/",parts.ToArray());
            }

            private void WriteProperty(XmlTextWriter elementsXml, string name, string value)
            {
                elementsXml.WriteStartElement("Property", nsSharePoint);
                elementsXml.WriteAttributeString("Name", name);
                elementsXml.WriteAttributeString("Value", value);
                elementsXml.WriteEndElement();
            }
            private void AddElementManifest(string fileName)
            {
                _FeatureXml.WriteStartElement("ElementManifest", nsSharePoint);
                _FeatureXml.WriteAttributeString("Location", fileName);
                _FeatureXml.WriteEndElement();
            }
            private void AddElementFile(string fileName)
            {
                _FeatureXml.WriteStartElement("ElementFile");
                _FeatureXml.WriteAttributeString("Location", fileName);
                _FeatureXml.WriteEndElement();
            }

            private string HandlePreview(List<string> _previewUrls, string previewUrl)
            {
                SPFile file = _WCMexportLib._RootWeb.GetFile(previewUrl);
                string fileName = RemoveUnderscore(file.Name);

                // Don't copy file if we've already done that
                //
                if (!_PreviousPreviewUrls.ContainsKey(fileName))
                {
                    _PreviousPreviewUrls.Add(fileName, true);
                    WriteToFile(file, Path.Combine(Path.Combine(_Dir, "MasterPages"), fileName));
                }

                // Don't add the Preview Image to the Feature.xml and "Elements".xml here
                // It's done in the dispose method to give a better structure
                //
                _previewUrls.Add(fileName);

                return String.Format("~SiteCollection/_catalogs/masterpage/Preview Images/{0}", fileName);
            }

            private string RemoveUnderscore(string str)
            {
                if (str.StartsWith("_"))
                    return str.Substring(1);
                return str;
            }

            private static string GetWebPartXml(Microsoft.SharePoint.WebPartPages.SPLimitedWebPartManager lwpm, WebPart wp)
            {
                StringWriter sw = new StringWriter();
                using (XmlTextWriter xw = new XmlTextWriter(sw))
                {
                    xw.Formatting = Formatting.Indented;
                    lwpm.ExportWebPart(wp, xw);
                }
                string wpExport = sw.ToString();

                // Replace CDATA sections with encoded content 
                // as nested CDATA sections aren't valid
                //
                Regex replaceCDATA = new Regex(@"<!\[CDATA\[(.*)?]]>",RegexOptions.Singleline);
                wpExport = replaceCDATA.Replace(wpExport, delegate(Match m){return HttpUtility.HtmlEncode(m.Groups[1].Value);});

                return wpExport;
            }

            public static void WriteToFile(SPFile spFile, string fileName)
            {
                using (Stream input = spFile.OpenBinaryStream())
                {
                    using (Stream output = File.Create(fileName))
                    {
                        byte[] buffer = new byte[32768];
                        while (true)
                        {
                            int read = input.Read(buffer, 0, buffer.Length);
                            if (read <= 0)
                                return;
                            output.Write(buffer, 0, read);
                        }
                    }
                }
            }

            public static string GetMD5hash(SPFile file)
            {
                MD5CryptoServiceProvider md5 = new MD5CryptoServiceProvider();
                using (Stream input = file.OpenBinaryStream())
                {
                    md5.ComputeHash(input);
                }
                StringBuilder key = new StringBuilder();
                foreach (byte hashByte in md5.Hash)
                {
                    key.Append(String.Format("{0:X1}", hashByte));
                }
                return key.ToString();
            }

            public void Dispose()
            {
                Dispose(true);
                GC.SuppressFinalize(this);
            }

            protected virtual void Dispose(bool disposing)
            {
                if (disposing)
                {
                    // Add the Preview Image to the Feature.xml and "Elements".xml
                    //
                    _PreviousPreviewUrls.Clear();
                    ProcessPreviewImageList(_MasterPagePreviewUrls, _MasterPagesXml, "MasterPagesPreviewImages");
                    ProcessPreviewImageList(_PageLayoutPreviewUrls, _PageLayoutsXml, "PageLayoutsPreviewImages");
                    if (_FeatureXml != null)
                    {
                        _FeatureXml.Close();
                        _FeatureXml = null;
                    }
                    if (_MasterPagesXml != null)
                    {
                        _MasterPagesXml.Close();
                        _MasterPagesXml = null;
                    }
                    if (_PageLayoutsXml != null)
                    {
                        _PageLayoutsXml.Close();
                        _PageLayoutsXml = null;
                    }
                    if (_StyleLibraryFilesXml != null)
                    {
                        _StyleLibraryFilesXml.Close();
                        _StyleLibraryFilesXml = null;
                    }
                }
            }
            private void ProcessPreviewImageList(List<string> previewUrls, XmlTextWriter elementXml, string moduleName)
            {
                if (previewUrls.Count > 0)
                {
                    // Close previous Module element
                    //
                    elementXml.WriteEndElement(); 

                    elementXml.WriteStartElement("Module", nsSharePoint);
                    elementXml.WriteAttributeString("Name", moduleName);
                    elementXml.WriteAttributeString("Url", "_catalogs/masterpage/Preview Images");
                    elementXml.WriteAttributeString("Path", "MasterPages");
                    elementXml.WriteAttributeString("RootWebOnly", "TRUE");
                    foreach (string fileName in previewUrls)
                    {
                        // Skip if already dealt with file
                        //
                        if (!_PreviousPreviewUrls.ContainsKey(fileName))
                        {
                            _PreviousPreviewUrls.Add(fileName, true);
                            AddElementFile(Path.Combine("MasterPages", fileName));

                            elementXml.WriteStartElement("File");
                            elementXml.WriteAttributeString("Url", fileName);
                            elementXml.WriteAttributeString("Type", "GhostableInLibrary");
                            elementXml.WriteEndElement();
                        }
                    }
                }
            }
        }
    }

    #region     public class UrlInfo
    public class UrlInfo
    {
        private string _Title;
        public string Title
        {
            get { return _Title; }
        }
        private string _Url;
        public string Url
        {
            get { return _Url; }
        }
        List<string> _Info = new List<string>();
        public ReadOnlyCollection<string> Info
        {
            get
            {
                return new ReadOnlyCollection<string>(_Info);
            }
        }
        public string InfoString
        {
            get
            {
                return string.Join("\r\n",_Info.ToArray());
            }
        }

        internal UrlInfo(string title, string url, string info)
        {
            _Title = title;
            _Url = url;
            _Info.Add(info);
        }

        internal void AddInfo(string info)
        {
            _Info.Add(info);
        }
    }
    #endregion

}
